/**
 * geis_dbus.c
 */
#include "geis_dbus.h"

#include <dbus/dbus.h>
#include "geis_debug.h"
#include <stdio.h>

#define GEIS_DBUS_MAX_BUS_NAME_LEN   64
#define GEIS_DBUS_MAX_BUS_OBJECT_LEN 64

#define GEIS_DBUS_BUS_NAME "com.canonical.geis"
#define GEIS_DBUS_INTERFACE_NAME "com.canonical.geis"
#define GEIS_DBUS_PATH "/com/canonical/geis"
static const char s_geis_match_gesture_started[] =
	"interface='" GEIS_DBUS_INTERFACE_NAME "',"
	"type='signal',"
	"path='" GEIS_DBUS_PATH "'";

struct _GeisDbus
{
  DBusConnection *connection;
  int             fd;
  char            busname[GEIS_DBUS_MAX_BUS_NAME_LEN];
  char            busobject[GEIS_DBUS_MAX_BUS_OBJECT_LEN];
};

static GeisDbus s_geis_dbus =
{
  NULL,
  -1,
  "",
  ""
};


static DBusHandlerResult
disconnect_message_handler(DBusConnection *connection,
                          DBusMessage    *message,
                          void           *data)
{
  if (dbus_message_is_signal(message, DBUS_INTERFACE_LOCAL, "Disconnected"))
  {
    geis_debug("disconnected from system bus.\n");
    s_geis_dbus.connection = NULL;
    return DBUS_HANDLER_RESULT_HANDLED;
  }

  return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}


static DBusHandlerResult
rpc_message_handler(DBusConnection *connection,
                    DBusMessage    *message,
                    void           *data)
{
  DBusHandlerResult result = DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
  DBusError error;
  dbus_error_init(&error);

  if (dbus_message_is_signal(message,
                             GEIS_DBUS_INTERFACE_NAME,
                             "gesture_started"))
  {
    geis_debug("gesture_started received\n");
    result = DBUS_HANDLER_RESULT_HANDLED;
  }
  else if (dbus_message_is_signal(message,
                                  GEIS_DBUS_INTERFACE_NAME,
                                  "gesture_updated"))
  {
    geis_debug("gesture_updated received\n");
    result = DBUS_HANDLER_RESULT_HANDLED;
  }
  else if (dbus_message_is_signal(message,
                                  GEIS_DBUS_INTERFACE_NAME,
                                  "gesture_finished"))
  {
    geis_debug("gesture_finished received\n");
    result = DBUS_HANDLER_RESULT_HANDLED;
  }

  dbus_error_free(&error);
  return result;
}


GeisDbus *
geis_dbus_get()
{
  if (!s_geis_dbus.connection)
  {
    DBusError error;
    dbus_error_init(&error);

    s_geis_dbus.connection = dbus_bus_get(DBUS_BUS_SYSTEM, &error);
    if (!s_geis_dbus.connection || dbus_error_is_set(&error))
    {
      geis_debug("error connecting to system bus: %s (%s)\n",
                 error.name, error.message);
      dbus_error_free(&error);
      return NULL;
    }

    dbus_connection_set_exit_on_disconnect(s_geis_dbus.connection, FALSE);

    if (!dbus_connection_get_unix_fd(s_geis_dbus.connection,
				     &s_geis_dbus.fd))
    {
      geis_debug("could not get descriptor for system bus.\n");
      dbus_error_free(&error);
      dbus_connection_unref(s_geis_dbus.connection);
      s_geis_dbus.connection = NULL;
      return NULL;
    }

    if (!dbus_connection_add_filter(s_geis_dbus.connection,
				    disconnect_message_handler,
				    NULL, NULL))
    {
      geis_debug("error adding disconnect filter: %s (%s)\n",
                 error.name, error.message);
      dbus_error_free(&error);
      dbus_connection_unref(s_geis_dbus.connection);
      s_geis_dbus.connection = NULL;
      return NULL;
    }

#if 0
    snprintf(s_geis_dbus.busname, sizeof(s_geis_dbus.busname), GEIS_DBUS_BUS_NAME);
    dbus_bus_request_name(s_geis_dbus.connection,
											    s_geis_dbus.busname,
											    0, &error);
    if (dbus_error_is_set(&error))
    {
	    geis_debug("error requesting %s: %s (%s)\n",
						     s_geis_dbus.busname, error.name, error.message);
	    dbus_error_free(&error);
	    dbus_connection_unref(s_geis_dbus.connection);
	    s_geis_dbus.connection = NULL;
	    return NULL;
    }
#endif

    snprintf(s_geis_dbus.busobject, sizeof(s_geis_dbus.busobject),
				     "/com/canonical/geis");
    dbus_bus_add_match(s_geis_dbus.connection,
		       s_geis_match_gesture_started,
		       &error);
    if (dbus_error_is_set(&error))
    {
      geis_debug("error adding match: %s (%s)\n", error.name, error.message);
      dbus_error_free(&error);
      dbus_connection_unref(s_geis_dbus.connection);
      s_geis_dbus.connection = NULL;
      return NULL;
    }

    DBusObjectPathVTable vtable = { .message_function = rpc_message_handler, };
    if (!dbus_connection_register_object_path(s_geis_dbus.connection,
					      s_geis_dbus.busobject,
					      &vtable,
					      NULL))
    {
	    geis_debug("error registering object path\n");
	    dbus_error_free(&error);
	    dbus_connection_unref(s_geis_dbus.connection);
	    s_geis_dbus.connection = NULL;
	    return NULL;
    }

#if 0
    geis_debug("dbus registered %s %s\n",
					      s_geis_dbus.busname, 
					      s_geis_dbus.busobject); 
#endif

    dbus_error_free(&error);
  }

  return &s_geis_dbus;
}


int
geis_dbus_fd(GeisDbus *dbus)
{
	return dbus->fd;
}


void
geis_dbus_dispatch(GeisDbus* dbus)
{
  if (s_geis_dbus.connection)
  {
    do
    {
      dbus_connection_read_write_dispatch(s_geis_dbus.connection, 0);
    } while (s_geis_dbus.connection
         && dbus_connection_get_dispatch_status(s_geis_dbus.connection)
             == DBUS_DISPATCH_DATA_REMAINS);
  }
}


void
geis_dbus_free(GeisDbus* dbus)
{
	dbus_connection_unref(dbus->connection);
}

