/**
 * geis_xcb.c
 *
 * Copyright 2010 Canonical Ltd.
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.	 See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */
#include "geis_xcb.h"

#include "geis_debug.h"
#include "geis_instance.h"
#include <grail.h>
#include <grail-types.h>
#include <stdlib.h>
#include <string.h>
#include <X11/X.h>
#include "xcb_gesture.h"
#include <xcb/xcb.h>

struct _GeisXcb
{
  GeisXcbWinInfo   *win_info;
  xcb_connection_t *connection;
  int               fd;
};

static GeisXcb s_geis_xcb =
{
  NULL,
  NULL,
  -1
};

typedef struct _GrailAttrMap
{
  const char *name;
  int         attr;
} GrailAttrMap;

static const GrailAttrMap s_grail_drag_attrs[] =
{
  { "delta X",    GRAIL_PROP_DRAG_DX }, 
  { "delta Y",    GRAIL_PROP_DRAG_DY },
  { "velocity X", GRAIL_PROP_DRAG_VX },
  { "velocity Y", GRAIL_PROP_DRAG_VY },
  { "position X", GRAIL_PROP_DRAG_X },
  { "position Y", GRAIL_PROP_DRAG_Y },
  { 0, 0 }
};

static const GrailAttrMap s_grail_pinch_attrs[] =
{
  { "radius delta",    GRAIL_PROP_PINCH_DR },
  { "radial velocity", GRAIL_PROP_PINCH_VR },
  { "radius",          GRAIL_PROP_PINCH_R },
  { "bounding box x1", GRAIL_PROP_PINCH_X1 },
  { "bounding box y1", GRAIL_PROP_PINCH_Y1 },
  { "bounding box x2", GRAIL_PROP_PINCH_X2 },
  { "bounding box y2", GRAIL_PROP_PINCH_Y2 },
 { 0, 0 }
};

static const GrailAttrMap s_grail_rotate_attrs[] =
{
  { "angle delta",      GRAIL_PROP_ROTATE_DA },
  { "angular velocity", GRAIL_PROP_ROTATE_VA },
  { "angle",            GRAIL_PROP_ROTATE_A },
  { 0, 0 }
};

static const GrailAttrMap s_grail_tap_attrs[] =
{
  { "tap time",   GRAIL_PROP_TAP_DT },
  { "position X", GRAIL_PROP_TAP_X },
  { "position Y", GRAIL_PROP_TAP_Y },
  { 0, 0 }
};

typedef struct _GrailTypeMap
{
  const char         *name;
  int                 type;
  const GrailAttrMap *attrs;
} GrailTypeMap;

static const GrailTypeMap s_grail_type_map[] =
{
  { "drag1",    GRAIL_TYPE_DRAG1,   s_grail_drag_attrs   },
  { "pinch1",   GRAIL_TYPE_PINCH1,  s_grail_pinch_attrs  },
  { "rotate1",  GRAIL_TYPE_ROTATE1, s_grail_rotate_attrs },
  { "tap1",     GRAIL_TYPE_TAP1,    s_grail_tap_attrs    },
  { "drag2",    GRAIL_TYPE_DRAG2,   s_grail_drag_attrs   },
  { "pinch2",   GRAIL_TYPE_PINCH2 , s_grail_pinch_attrs  },
  { "rotate2",  GRAIL_TYPE_ROTATE2, s_grail_rotate_attrs },
  { "tap2",     GRAIL_TYPE_TAP2,    s_grail_tap_attrs    },
  { "drag3",    GRAIL_TYPE_DRAG3,   s_grail_drag_attrs   },
  { "pinch3",   GRAIL_TYPE_PINCH3,  s_grail_pinch_attrs  },
  { "rotate3",  GRAIL_TYPE_ROTATE3, s_grail_rotate_attrs },
  { "tap3",     GRAIL_TYPE_TAP3,    s_grail_tap_attrs    },
  { "drag4",    GRAIL_TYPE_DRAG4,   s_grail_drag_attrs   },
  { "pinch4",   GRAIL_TYPE_PINCH4,  s_grail_pinch_attrs  },
  { "rotate4",  GRAIL_TYPE_ROTATE4, s_grail_rotate_attrs },
  { "tap4",     GRAIL_TYPE_TAP4,    s_grail_tap_attrs    },
  { "drag5",    GRAIL_TYPE_DRAG5,   s_grail_drag_attrs   },
  { "pinch5",   GRAIL_TYPE_PINCH5,  s_grail_pinch_attrs  },
  { "rotate5",  GRAIL_TYPE_ROTATE5, s_grail_rotate_attrs },
  { "tap5",     GRAIL_TYPE_TAP5,    s_grail_tap_attrs    },
  { "edrag",    GRAIL_TYPE_EDRAG,   s_grail_drag_attrs   },
  { "epinch",   GRAIL_TYPE_EPINCH,  s_grail_pinch_attrs  },
  { "erotate",  GRAIL_TYPE_EROTATE, s_grail_rotate_attrs },
  { "mdrag",    GRAIL_TYPE_MDRAG,   s_grail_drag_attrs   },
  { "mpinch",   GRAIL_TYPE_MPINCH,  s_grail_pinch_attrs  },
  { "mrotate",  GRAIL_TYPE_MROTATE, s_grail_rotate_attrs },
  { 0, 0, 0 }
};


GeisGestureAttr s_name_attr =
{
  .name="name",
  .type=GEIS_ATTR_TYPE_STRING
};

/**
 * Dispatches gesture events to the instance's callback.
 */
static void
geis_xcb_dispatch_gesture_type(GeisInstance instance,
                               const char **gesture_list)
{
  if (gesture_list == GEIS_ALL_GESTURES)
  {
      const GrailTypeMap *type = s_grail_type_map;
      for (; type->name; ++type)
      {
	s_name_attr.string_val = type->name;
	if (instance->gesture_funcs &&
	    instance->gesture_funcs->added)
	{
	  instance->gesture_funcs->added (instance->gesture_cookie,
					  type->type,
					  0,
					  1, &s_name_attr);
	}
      }
  }
  else
  {
    const char **gesture = NULL;
    for (gesture = gesture_list; *gesture; ++gesture)
    {
      const GrailTypeMap *type = s_grail_type_map;
      for (; type->name; ++type)
      {
      	if (strcmp(type->name, *gesture) == 0)
	{
	  s_name_attr.string_val = type->name;
	  if (instance->gesture_funcs && instance->gesture_funcs->added)
	  {
	    instance->gesture_funcs->added (instance->gesture_cookie,
					    type->type,
					    0,
					    1, &s_name_attr);
	  }

	  break;
	}
      }
    }
  }
}


/**
 * Dispatches gesture events to the instance's callback.
 */
static void
geis_xcb_dispatch_gesture(GeisInstance                instance,
                          xcb_gesture_notify_event_t *event)
{
  static const int num_standard_attrs = 8;
  int num_attrs = num_standard_attrs + event->num_props;
  int i = 0;
  float *properties = NULL;

  GeisGestureAttr *attrs = calloc(num_attrs, sizeof(GeisGestureAttr));
  if (!attrs)
  {
    return;
  }

  attrs[0].name = "device id";
  attrs[0].type = GEIS_ATTR_TYPE_INTEGER;
  attrs[0].integer_val = event->device_id;
  attrs[1].name = "timestamp";
  attrs[1].type = GEIS_ATTR_TYPE_INTEGER;
  attrs[1].integer_val = event->time;
  attrs[2].name = "root window id";
  attrs[2].type = GEIS_ATTR_TYPE_INTEGER;
  attrs[2].integer_val = event->root;
  attrs[3].name = "event window id";
  attrs[3].type = GEIS_ATTR_TYPE_INTEGER;
  attrs[3].integer_val = event->event;
  attrs[4].name = "child window id";
  attrs[4].type = GEIS_ATTR_TYPE_INTEGER;
  attrs[4].integer_val = event->child;
  attrs[5].name = "focus X";
  attrs[5].type = GEIS_ATTR_TYPE_FLOAT;
  attrs[5].float_val = event->focus_x;
  attrs[6].name = "focus Y";
  attrs[6].type = GEIS_ATTR_TYPE_FLOAT;
  attrs[6].float_val = event->focus_y;

  {
    const GrailTypeMap *type = s_grail_type_map;
    properties = (float *)(event + 1);

    attrs[7].name = "name";
    attrs[7].type = GEIS_ATTR_TYPE_STRING;
    attrs[7].string_val = "unknown";
    for (; type->name; ++type)
    {
      if (type->type == event->gesture_type)
      {
	int i;
	attrs[7].string_val = type->name;

	for (i=0; i < event->num_props; ++i)
	{
	  attrs[num_standard_attrs + i].name = type->attrs[i].name;
	  attrs[num_standard_attrs + i].type = GEIS_ATTR_TYPE_FLOAT;
	  attrs[num_standard_attrs + i].float_val = properties[i];
	}

	break;
      }
    }
  }

  switch (event->status)
  {
    case GRAIL_STATUS_BEGIN:
      if (instance->gesture_funcs &&
	  instance->gesture_funcs->start)
      {
	instance->gesture_funcs->start (instance->gesture_cookie,
					event->gesture_type,
					event->gesture_id,
					num_attrs, attrs);
      }

      break;
    case GRAIL_STATUS_UPDATE:
      if (instance->gesture_funcs &&
	  instance->gesture_funcs->update)
      {
	instance->gesture_funcs->update (instance->gesture_cookie,
					 event->gesture_type,
					 event->gesture_id,
					 num_attrs, attrs);
      }

      break;
    case GRAIL_STATUS_END:
      if (instance->gesture_funcs &&
	  instance->gesture_funcs->finish)
      {
	instance->gesture_funcs->finish (instance->gesture_cookie,
					 event->gesture_type,
					 event->gesture_id,
					 num_attrs, attrs);
      }

      break;
  }

  free(attrs);
}


/**
 * @param[in] win_info
 */
GeisXcb *
geis_xcb_get(GeisXcbWinInfo *win_info)
{
  if (!s_geis_xcb.connection)
  {
    xcb_generic_error_t *error;

    s_geis_xcb.win_info = win_info;
    s_geis_xcb.connection = xcb_connect(win_info->display_name,
                                        win_info->screenp);
    if (!s_geis_xcb.connection)
    {
      geis_debug("error connecting to X server.");
      return NULL;
    }

#if 0
    {
      char buffer[128];
      xcb_input_get_extension_version_cookie_t input_version_cookie;

      input_version_cookie = xcb_input_get_extension_version(s_geis_xcb.connection,
                                                             127,
                                                             buffer);
    }
#endif

    {
      xcb_gesture_query_version_cookie_t  version_cookie;
      xcb_gesture_query_version_reply_t  *version_reply = NULL;

      version_cookie = xcb_gesture_query_version(s_geis_xcb.connection,
                                                 XCB_GESTURE_MAJOR_VERSION,
                                                 XCB_GESTURE_MINOR_VERSION);
      version_reply = xcb_gesture_query_version_reply(s_geis_xcb.connection,
                                                      version_cookie,
                                                      &error);
      if (!version_reply)
      {
	geis_debug("failed to receive gesture version reply\n");
	return NULL;
      }

      if (version_reply->major_version != XCB_GESTURE_MAJOR_VERSION
       && version_reply->minor_version != XCB_GESTURE_MINOR_VERSION)
      {
	geis_debug("server supports unrecognized version: %d.%d\n",
	           version_reply->major_version, version_reply->minor_version);
	return NULL;
      }

     free(version_reply);
    }


    s_geis_xcb.fd = xcb_get_file_descriptor(s_geis_xcb.connection);
  }

  return &s_geis_xcb;
}


int
geis_xcb_fd(GeisXcb *xcb)
{
	return xcb->fd;
}


void
geis_xcb_dispatch(GeisInstance instance)
{
  GeisXcb *xcb = instance->xcb;
  if (xcb->connection)
  {
    const xcb_query_extension_reply_t *extension_info;
    extension_info = xcb_get_extension_data(xcb->connection, &xcb_gesture_id);

    xcb_generic_event_t *event = xcb_poll_for_event(xcb->connection);
    while (event)
    {
      xcb_gesture_notify_event_t *gesture_event = NULL;
      if (event->response_type != GenericEvent) {
	geis_debug("warning: Received non-generic event type: %d\n",
	           event->response_type);
	continue;
      }

      gesture_event = (xcb_gesture_notify_event_t*)event;
      if (gesture_event->extension != extension_info->major_opcode)
      {
	geis_debug("warning: Received non-gesture extension event: %d\n",
	           gesture_event->extension);
	continue;
      }

      if (gesture_event->event_type != XCB_GESTURE_NOTIFY)
      {
	geis_debug("warning: Received unrecognized gesture event type: %d\n",
	           gesture_event->event_type);
	continue;
      }

      geis_xcb_dispatch_gesture(instance, gesture_event);

      event = xcb_poll_for_event(xcb->connection);
    }
  }
}


void
geis_xcb_free(GeisXcb* xcb)
{
  xcb_disconnect(xcb->connection);
}


GeisStatus
geis_xcb_subscribe(GeisInstance   instance,
                   const char   **gesture_list)
{
  uint16_t device_id = 0; //< temporary for dev testing
  uint32_t mask_len = 1; //< temporary for dev testing
  uint32_t mask[] = { 0xffffffff }; //< temporary for dev testing

  xcb_generic_error_t *error;
  xcb_void_cookie_t select_cookie;
  GeisXcb *xcb = instance->xcb;

  geis_xcb_dispatch_gesture_type(instance, gesture_list);

  select_cookie = xcb_gesture_select_events_checked(xcb->connection,
						    xcb->win_info->window_id,
						    device_id,
						    mask_len,
						    mask);
  error = xcb_request_check(xcb->connection, select_cookie);
  if (error)
  {
    geis_debug("failed to select events\n");
    return GEIS_UNKNOWN_ERROR;
  }

  {
    xcb_gesture_get_selected_events_cookie_t  events_cookie;
    xcb_gesture_get_selected_events_reply_t  *events_reply = NULL;
    xcb_gesture_event_mask_iterator_t         events_iterator;
    unsigned masks_len_reply = 0;
    unsigned mask_len_reply = 0;
    uint32_t *mask_reply;

    events_cookie = xcb_gesture_get_selected_events(xcb->connection,
						    xcb->win_info->window_id);
    events_reply = xcb_gesture_get_selected_events_reply(xcb->connection,
							 events_cookie,
							 &error);
    if (!events_reply)
    {
      geis_debug("failed to receive selected events reply\n");
      return GEIS_UNKNOWN_ERROR;
    }

    masks_len_reply = xcb_gesture_get_selected_events_masks_length(events_reply);
    if (masks_len_reply != 1)
    {
      geis_debug("wrong selected masks length returned: %d\n", masks_len_reply);
      return GEIS_UNKNOWN_ERROR;
    }

    events_iterator = xcb_gesture_get_selected_events_masks_iterator(events_reply);
    if (events_iterator.data->device_id != device_id)
    {
      geis_debug("incorrect device id returned by server: %d\n",
                 events_iterator.data->device_id);
      return GEIS_UNKNOWN_ERROR;
    }

    mask_len_reply = xcb_gesture_event_mask_mask_data_length(events_iterator.data);
    if (mask_len_reply != mask_len)
    {
      geis_debug("incorrect mask length returned by server: %d\n",
                 mask_len_reply);
      return GEIS_UNKNOWN_ERROR;
    }

    mask_reply = xcb_gesture_event_mask_mask_data(events_iterator.data);
    if (memcmp(mask, mask_reply, mask_len * 4) != 0)
    {
      geis_debug("incorrect mask returned by server\n");
      return GEIS_UNKNOWN_ERROR;
    }

    free(events_reply);
  }

  return GEIS_STATUS_SUCCESS;
}
